use std::pin::Pin;

use tokio_stream::{
    wrappers::{ReceiverStream, WatchStream},
    Stream, StreamExt,
};
use tonic::{Request, Response, Status};

use crate::system::health::HealthMonitor;

use super::proto::*;

type AgentResult<T> = std::result::Result<Response<T>, Status>;

pub struct AgentService {
    health: HealthMonitor,
}

#[tonic::async_trait]
impl agent_server::Agent for AgentService {
    type HealthStream = Pin<Box<dyn Stream<Item = Result<HealthResponse, Status>> + Send>>;

    async fn health(&self, _: Request<()>) -> AgentResult<Self::HealthStream> {
        let receiver = self.health.monitor();
        let version = env!("CARGO_PKG_VERSION");

        let output = WatchStream::new(receiver).map(|health| {
            Ok(HealthResponse {
                version: version.to_owned(),
                system: Some(health.system().into()),
                tasks: health
                    .tasks()
                    .into_iter()
                    .map(|(key, val)| (key, TaskHealth::from(val)))
                    .collect(),
            })
        });

        Ok(Response::new(Box::pin(output)))
    }

    async fn get_sys_info(&self, _: Request<()>) -> AgentResult<SysInfoResponse> {
        use sysinfo::{CpuExt, DiskExt, SystemExt};

        let mut sys = crate::system::SYSTEM.lock().unwrap();

        sys.refresh_specifics(
            sysinfo::RefreshKind::new()
                // .with_disks() // what is this?
                .with_disks_list()
                .with_memory()
                .with_processes(sysinfo::ProcessRefreshKind::everything())
                .with_cpu(sysinfo::CpuRefreshKind::everything()),
        );

        let cpus = sys
            .cpus()
            .iter()
            .map(|cpu| sys_info_response::Cpu {
                freq_mhz: cpu.frequency(),
                usage: cpu.cpu_usage(),
            })
            .collect();

        let disks = sys
            .disks()
            .iter()
            .map(|disk| sys_info_response::Disk {
                name: disk.name().to_string_lossy().into_owned(),
                total_bytes: disk.total_space(),
                avail_bytes: disk.available_space(),
                mount_point: disk.mount_point().to_string_lossy().into_owned(),
            })
            .collect();

        let response = Response::new(SysInfoResponse {
            uptime: sys.uptime(),
            hostname: sys.host_name().unwrap_or_default(),
            os: sys.long_os_version().unwrap_or_default(),
            mem_total_bytes: sys.total_memory(),
            mem_avail_bytes: sys.available_memory(),
            swap_total_bytes: sys.total_swap(),
            swap_free_bytes: sys.free_swap(),
            cpus,
            disks,
        });

        Ok(response)
    }

    type ExecStream = Pin<Box<dyn Stream<Item = Result<ExecResponse, Status>> + Send>>;

    async fn exec(&self, req: Request<ExecRequest>) -> AgentResult<Self::ExecStream> {
        use crate::system::exec::*;

        let ExecRequest { program, args } = req.get_ref();

        match exec(program, args) {
            Ok(receiver) => {
                let stream = ReceiverStream::new(receiver).map(|_inner| {
                    Ok(ExecResponse {
                        // TODO
                        response: None,
                    })
                });

                Ok(Response::new(Box::pin(stream)))
            }
            Err(err) => Err(Status::failed_precondition(err.to_string())),
        }
    }
}

pub fn server(health_monitor: HealthMonitor) -> agent_server::AgentServer<AgentService> {
    agent_server::AgentServer::new(AgentService {
        health: health_monitor,
    })
}
