use std::time::Duration;

use tokio::{signal, sync::oneshot, time::sleep};
use tower_http::trace::TraceLayer;

use crate::system::health::HealthMonitor;

mod agent;
mod proto {
    tonic::include_proto!("prymn");

    impl From<crate::system::health::SystemHealth> for SystemHealth {
        fn from(val: crate::system::health::SystemHealth) -> Self {
            if let crate::system::health::SystemStatus::Critical(ref reasons) = val.status {
                SystemHealth {
                    status: itertools::join(reasons.iter().map(ToString::to_string), ","),
                }
            } else {
                SystemHealth {
                    status: val.status.to_string(),
                }
            }
        }
    }

    impl From<crate::system::health::TaskHealth> for TaskHealth {
        fn from(val: crate::system::health::TaskHealth) -> Self {
            TaskHealth {
                status: val.status().to_string(),
                message: val.message().to_owned(),
                started_on: val.started_on().to_string(),
                progress: val.progress() as i32,
            }
        }
    }
}

/// Run the server. This is the main entry point of the application.
#[tokio::main]
pub async fn run() -> anyhow::Result<()> {
    let (shutdown_tx, shutdown_rx) = oneshot::channel();

    // Listen for shutdown signals
    tokio::spawn(async {
        signal::ctrl_c()
            .await
            .expect("failed to listen to a ctrl-c signal");

        let _ = shutdown_tx.send(());
    });

    let health_monitor = HealthMonitor::new();
    let agent_service = agent::server(health_monitor.clone());

    tokio::spawn(async move {
        loop {
            health_monitor.check_system().await;
            sleep(Duration::from_secs(1)).await;
        }
    });

    let addr = "[::]:50012".parse()?;
    tracing::info!("listening on {}", addr);
    tonic::transport::Server::builder()
        .layer(TraceLayer::new_for_grpc())
        .add_service(agent_service)
        .serve_with_shutdown(addr, async {
            let _ = shutdown_rx.await;
        })
        .await?;

    Ok(())
}
